using System.Net;
using FluentAssertions;
using JsonApiDotNetCore.Serialization.Objects;
using Microsoft.EntityFrameworkCore;
using TestBuildingBlocks;
using Xunit;

namespace JsonApiDotNetCoreTests.IntegrationTests.AtomicOperations.Transactions;

public sealed class AtomicRollbackTests : IClassFixture<IntegrationTestContext<TestableStartup<OperationsDbContext>, OperationsDbContext>>
{
    private readonly IntegrationTestContext<TestableStartup<OperationsDbContext>, OperationsDbContext> _testContext;
    private readonly OperationsFakers _fakers = new();

    public AtomicRollbackTests(IntegrationTestContext<TestableStartup<OperationsDbContext>, OperationsDbContext> testContext)
    {
        _testContext = testContext;

        testContext.UseController<OperationsController>();
    }

    [Fact]
    public async Task Can_rollback_on_error()
    {
        // Arrange
        string newArtistName = _fakers.Performer.Generate().ArtistName!;
        DateTimeOffset newBornAt = _fakers.Performer.Generate().BornAt;
        string newTitle = _fakers.MusicTrack.Generate().Title;

        await _testContext.RunOnDatabaseAsync(async dbContext =>
        {
            await dbContext.ClearTablesAsync<Performer, MusicTrack>();
        });

        string unknownPerformerId = Unknown.StringId.For<Performer, int>();

        var requestBody = new
        {
            atomic__operations = new object[]
            {
                new
                {
                    op = "add",
                    data = new
                    {
                        type = "performers",
                        attributes = new
                        {
                            artistName = newArtistName,
                            bornAt = newBornAt
                        }
                    }
                },
                new
                {
                    op = "add",
                    data = new
                    {
                        type = "musicTracks",
                        attributes = new
                        {
                            title = newTitle
                        },
                        relationships = new
                        {
                            performers = new
                            {
                                data = new[]
                                {
                                    new
                                    {
                                        type = "performers",
                                        id = unknownPerformerId
                                    }
                                }
                            }
                        }
                    }
                }
            }
        };

        const string route = "/operations";

        // Act
        (HttpResponseMessage httpResponse, Document responseDocument) = await _testContext.ExecutePostAtomicAsync<Document>(route, requestBody);

        // Assert
        httpResponse.Should().HaveStatusCode(HttpStatusCode.NotFound);

        responseDocument.Errors.ShouldHaveCount(1);

        ErrorObject error = responseDocument.Errors[0];
        error.StatusCode.Should().Be(HttpStatusCode.NotFound);
        error.Title.Should().Be("A related resource does not exist.");
        error.Detail.Should().Be($"Related resource of type 'performers' with ID '{unknownPerformerId}' in relationship 'performers' does not exist.");
        error.Source.ShouldNotBeNull();
        error.Source.Pointer.Should().Be("/atomic:operations[1]");

        await _testContext.RunOnDatabaseAsync(async dbContext =>
        {
            List<Performer> performersInDatabase = await dbContext.Performers.ToListAsync();
            performersInDatabase.Should().BeEmpty();

            List<MusicTrack> tracksInDatabase = await dbContext.MusicTracks.ToListAsync();
            tracksInDatabase.Should().BeEmpty();
        });
    }

    [Fact]
    public async Task Can_restore_to_previous_savepoint_on_error()
    {
        // Arrange
        string newTrackTitle = _fakers.MusicTrack.Generate().Title;

        await _testContext.RunOnDatabaseAsync(async dbContext =>
        {
            await dbContext.ClearTablesAsync<Performer, MusicTrack>();
        });

        const string trackLid = "track-1";

        string unknownPerformerId = Unknown.StringId.For<Performer, int>();

        var requestBody = new
        {
            atomic__operations = new object[]
            {
                new
                {
                    op = "add",
                    data = new
                    {
                        type = "musicTracks",
                        lid = trackLid,
                        attributes = new
                        {
                            title = newTrackTitle
                        }
                    }
                },
                new
                {
                    op = "add",
                    @ref = new
                    {
                        type = "musicTracks",
                        lid = trackLid,
                        relationship = "performers"
                    },
                    data = new[]
                    {
                        new
                        {
                            type = "performers",
                            id = unknownPerformerId
                        }
                    }
                }
            }
        };

        const string route = "/operations";

        // Act
        (HttpResponseMessage httpResponse, Document responseDocument) = await _testContext.ExecutePostAtomicAsync<Document>(route, requestBody);

        // Assert
        httpResponse.Should().HaveStatusCode(HttpStatusCode.NotFound);

        responseDocument.Errors.ShouldHaveCount(1);

        ErrorObject error = responseDocument.Errors[0];
        error.StatusCode.Should().Be(HttpStatusCode.NotFound);
        error.Title.Should().Be("A related resource does not exist.");
        error.Detail.Should().Be($"Related resource of type 'performers' with ID '{unknownPerformerId}' in relationship 'performers' does not exist.");
        error.Source.ShouldNotBeNull();
        error.Source.Pointer.Should().Be("/atomic:operations[1]");

        await _testContext.RunOnDatabaseAsync(async dbContext =>
        {
            List<Performer> performersInDatabase = await dbContext.Performers.ToListAsync();
            performersInDatabase.Should().BeEmpty();

            List<MusicTrack> tracksInDatabase = await dbContext.MusicTracks.ToListAsync();
            tracksInDatabase.Should().BeEmpty();
        });
    }
}
